#ifndef QUCIRCULARPLOTATTRIBUTES_H
#define QUCIRCULARPLOTATTRIBUTES_H

#include <QColor>

class OOBDistort;

struct  plot_axis_x {
    float baseline = 0.0f;
    bool autoscale = true;
    double lb = -1000;
    double ub = 1000;  // "axis" bounds, either manual or auto
};

struct plot_axis_y {
    bool autoscale = true;
    bool scale_inverted = false;
    double lb = -1000;
    double ub = 1000;
};

struct plot_axes {
    struct plot_axis_x x;
    struct plot_axis_y y;
};

struct plot_selection {
    float aradius = 3.5; //!< radius of the (moving) circular selection area
    float yp_radius = 1.2; //!< radius of the point selected on the curve
}; //!< selection attributes

struct  plot_geom {
    double inner_radius_factor = 0.2; //!< where (geometrically) the lower bound is
    double outer_radius_factor = 0.8; //!< where (geometrically) the upper bound is
    OOBDistort *oob_distortion = nullptr; //!< "out of bounds" curve distortion
        //!< defines how curves are distorted when points
        //!< exceed Y axis upper and lower bounds
        //!< Default must be initialized (QuCircularPlotEngine)
    struct plot_selection selection;
}; //!< geometry attributes: inner and outer radius, radius factor

struct plot_colors {
    QColor background;   //!<  background color. Default: QPalette::Window
    QColor axes;  //!<  axes color. Default: QPalette::WindowText
    QColor selection; //!< color of the selected point. Default: QPalette::WindowText.darker()
};

struct  plot_palette {
    QList<QColor> dark = QList<QColor>() <<
                         QColor(25, 229, 107) <<  // Green
                         QColor(229, 209, 25) <<  // Yellow
                         QColor(25, 107, 229) <<  // Blue
                         QColor(229, 147, 25) <<  // Orange
                         QColor(127, 25, 229) <<  // Purple
                         QColor(229, 25, 86) <<  // Brown
                         QColor(229, 86, 25) <<  // Brown
                         QColor(25, 229, 45) <<  // Green
                         QColor(229, 25, 25)  // Red
        ;

    QList<QColor> light = QList<QColor>() <<
                          QColor(127, 229, 25) <<  // Green
                          QColor(229, 127, 25) <<  // Orange
                          QColor(229, 229, 25) <<  // Yellow
                          QColor(25, 229, 127) <<  // Green
                          QColor(25, 229, 229) <<  // Cyan
                          QColor(25, 127, 229) <<  // Blue
                          QColor(229, 25, 25) <<  // Red
                          QColor(25, 25, 229) <<  // Blue
                          QColor(127, 25, 229) <<  // Purple
                          QColor(229, 25, 229) <<  // Magenta
                          QColor(229, 25, 127) <<  // Purple
                          QColor(25, 229, 25) // Green
        ;


    bool theme_dark = false;
};

struct plot_paint {
    bool show_values = false; //!<  show values on the plot. Default: false. OK when there's little data
    bool show_points = false; //!<  show data points (default: false, while show_curves is true by default)
    bool show_bounds = true; //!<  draw one circle on the lower bound and one on the upper (default: yes)

    bool show_curves = true;  //!< show the curves (lines connecting the data points): default true
    bool show_xax = true; //!<  show the "X" axis: default: yes on the baseline
    bool show_origin = true; //!< show the "origin" at 9 o'clock default: true

    bool close_curve = true; //!< close the curve, drawing a line between y[size-1] and y[0]

    int y_axes = 0; //!< number of "Y" axes to show default: 0

    struct plot_colors colors;
    struct plot_palette palette;

    bool debug = false;

}; //!< "paint" attributes: show or hide things

/*!
 * \brief Holds QuCircularPlot options
 *
 * This struct hierarchically stores the characteristics of the plot
 *
 * \section axes
 * Defines the axes configuration
 *
 * \subsection x
 * Defines the x axis configuration
 * - baseline
 * - autoscale
 * - lower bound
 * - upper bound
 *
 * \subsection y
 * Defines the y axis configuration:
 * - scale_inverted
 * - autoscale
 * - lower bound
 * - upper bound
 *
 * \section geom
 * Defines the geometrical characteristics of the plot: the inner and outer radius representing the
 * axis limits (lower and upper) and the kind of distortion that shall be applied to curves when
 * their points exceed the (non automatic) Y bounds.
 *
 * \subsection selection
 * Geometry for the selection (movable selection area and selected point on the curve)
 *
 * \section paint
 * Contains drawing options (show hide elements) and *colors* in the dedicated section
 *
 * \subsection colors
 * Define colors for several elements
 *
 * \section Note
 * Whenever you change one or more attributes, you may want to trigger an immediate *update*
 * on the plot (either item or widget)
 *
 * \section Example
 *
 * \code
    QuCircularPlotI *ni = new QuCircularPlotI(nullptr, QSize(300, 300));
    QuCircularPlotEngine *pie = ni->engine();
    qucircularplot_attributes &a = pie->a;
    a.axes.y.autoscale = true;  // enable y axis autoscale
    a.geom.inner_radius_factor = 0.1; // widen inter axes area
    a.paint.colors.background = QColor(Qt::white);  // light color scheme
    ni->update();
 * \endcode
 *
 */
struct qucircularplot_attributes
{
    struct plot_axes axes;
    struct plot_geom geom;
    struct plot_paint paint;
};

#endif // QUCIRCULARPLOTATTRIBUTES_H
