#include "qunumericengine.h"
#include <math.h>
#include <QPainter>
#include <QtDebug>
#include <QWidget>
#include <cumacros.h>
#include <QTimer>

#define CHECK_MARK "✓"

QuNumericEngine::QuNumericEngine(const QFont& f)
    : QObject{nullptr}  {
    data = new QuNumEngineData;
    data->intDig = 3;
    data->decDig = 1;
    data->digits = 4;
    data->data = data->last_applied_data =  0;
    data->min = (long long int) -pow(10.0, data->digits) + 1;
    data->max = (long long int) pow(10.0, data->digits) - 1;
    data->min_d = -pow(10.0, data->intDig) + pow(10.0, -data->decDig);
    data->max_d = pow(10.0, data->intDig) - pow(10.0, -data->decDig);
    data->font_scale = 2.0;
    data->font = f;
    data->font.setPointSizeF(f.pointSizeF() * data->font_scale);
    data->arrow_hei = QFontMetrics(data->font).height() / 4.0;
    data->lm = data->bm = data->um = data->rm = 0.0;
    data->mouse_down = false;
    data->show_apply = true;
    m_update_maxcharw();
}

QuNumericEngine::~QuNumericEngine() {
    delete data;
}

void QuNumericEngine::contextMenuEvent(const QPointF &pos)
{

}

void QuNumericEngine::mousePressEvent(const QPointF &pos) {
    data->mouse_pos = pos;
    data->mouse_down = true;
}

void QuNumericEngine::mouseMoveEvent(const QPointF &pos) {
    data->mouse_pos = pos;
}

void QuNumericEngine::mouseReleaseEvent(const QPointF &pos) {
    data->mouse_pos = pos;
    data->mouse_down = false;
    if(!pos.isNull())
        click(pos);
}

void QuNumericEngine::mouseDoubleClickEvent(const QPointF &pos)
{

}

void QuNumericEngine::wheelEvent(const QPointF &pos, double delta) {
    click(pos, delta > 0);
}

double QuNumericEngine::value() const {
    return data->data * pow(10.0, -data->decDig);
}

double QuNumericEngine::minimum() const {
    return data->min_d;
}

double QuNumericEngine::maximum() const {
    return data->max_d;
}

void QuNumericEngine::paint(QPainter *p, const QRectF &rect, QWidget *widget) {
    QColor bgcolor = widget->palette().color(QPalette::Window);
    QFont f = p->font();
    f.setPointSizeF(f.pointSizeF() * data->font_scale);
    p->setFont(f);
    p->setRenderHint(QPainter::Antialiasing, true);
    QFontMetrics fm(f);
    double fh =  fm.ascent() + fm.descent();
    data->digitrects.clear();

    double x = 0; // current digit position during loop
    long long temp = data->data;
    int num = 0;
    const QString& sign = (data->data < 0) ? "-" : "";
    // calculate rects for digits
    double digw, dotw; // widths for sign, digits and dot
    const double& w = rect.width(); // available width
    const double& h = rect.height();
    // estimate the proportion between the width needed for the dot and a normal digit
    // give the sign the same width as a digit
    double dotwr = fm.horizontalAdvance(".") / (double) fm.horizontalAdvance('8');
    double t = data->digits + (sign.length() > 0 ? 1 : 0) + (data->decDig > 0 ? dotwr : 0) + (data->show_apply ? 1 : 0);
    digw = w / t;
    dotw = digw * dotwr;

    // draw sign
    if(sign.length() > 0) {
        QuANRect tr(x, 0, digw, h, QuANRect::Sign);
        p->drawText(tr, Qt::AlignHCenter|Qt::AlignVCenter, "-");
        data->digitrects << tr;
        x += tr.width();
    }

    for (int i = 0; i < data->digits; i++)
    {
        if (i == data->intDig)  {
            QuANRect tr(x, 0, dotw, h, QuANRect::Dot);
            p->drawText(tr,  Qt::AlignHCenter|Qt::AlignVCenter, ".");
            data->digitrects << tr;
            x += dotw;
        }
        QuANRect tr(x, 0, digw, h, i), dr(tr);
        dr.setLeft(dr.left() + 2);
        dr.setTop(dr.top() + 2);
        dr.setBottom(dr.bottom() - 1);
        dr.setRight(dr.right() - 1);
        QRectF r(tr);
        p->setPen(Qt::black);
        num = (long long) (temp / (long long) pow(10.0, data->digits-i-1));
        temp -= num * (long) pow(10.0, data->digits-i-1);
        const QString& txt = QString("%1").arg(abs(num)); // draw after highlight
        tr.setRect(x, 0, digw, h);
        data->digitrects << tr;
        x += tr.width();
        if(dr.contains(data->mouse_pos)) {
            double x1, x2, y1, y2, h;
            double xoff = dr.width() / 10.0;
            QPen save_pen = p->pen();
            p->setPen(data->mouse_down ? Qt::darkGray : bgcolor.darker(115));
            QColor highlightC(data->mouse_down ? bgcolor.darker(110) : bgcolor.darker(70));
            QBrush savebrush = p->brush();
            p->setBrush(highlightC);
            if(data->mouse_pos.y() > dr.height() / 2.0) {
                p->drawRoundedRect(r.x(), r.y() + r.height() / 2.0, r.width(), r.height() / 2.0, 1.0, 1.0);
                // arrow down
                p->setPen(bgcolor.darker(180));
                x1 = dr.left() + xoff;
                x2 = x1 + (dr.width() - 2 * xoff) / 2.0;
                y1 = dr.bottom() - dr.height() / 2.0 + fh / 2.0 ;
                data->arrow_hei < (dr.height() - fh)/2.0 ? h = data->arrow_hei : h = (dr.height() - fh)/2.0;
                y2 = y1 + h;
                p->drawLine(x1, y1, x2, y2);
                x1 = x2;
                y1 = y2;
                x2 = dr.right() - xoff;
                y2 = y1 - h;
                p->drawLine(x1, y1, x2, y2);
            }
            else {
                p->drawRoundedRect(r.x(), r.y(), r.width(), r.height() / 2.0, 1.0, 1.0);
                // arrow up
                p->setPen(bgcolor.darker(180));
                x1 = dr.left() + xoff;
                x2 = x1 + (dr.width() - 2 * xoff) / 2.0;
                data->arrow_hei < (dr.height() - fh)/2.0 ? h = data->arrow_hei : h = (dr.height() - fh)/2.0;
                y1 = dr.top() + dr.height() / 2.0 - fh / 2.0;
                y2 = y1 - h;
                p->drawLine(x1, y1, x2, y2);
                x1 = x2;
                y1 = y2;
                x2 = dr.right() - xoff;
                y2 = y1 + h;
                p->drawLine(x1, y1, x2, y2);
            }
            p->setPen(save_pen);
            p->setBrush(savebrush);
        }


        //        QRectF br; // text bounding rect
        p->drawText(tr,  Qt::AlignHCenter|Qt::AlignVCenter, txt/*, &br*/);
        //        p->setPen(Qt::cyan);
        //        p->drawRect(br);

        //
        QPen pe = p->pen();
        pe.setColor(bgcolor);
        pe.setWidthF(fh / 30);
        p->setPen(pe);
        double lw = /*1.1 **/ data->maxcharw, ls = tr.left() + (tr.width() - lw) / 2.0;
        p->drawLine(ls, tr.height()/2.0, ls + lw, tr.height() / 2.0);
        p->setPen(Qt::lightGray);
        p->drawRect(data->digitrects.last());
    }
    if(data->show_apply) {
        p->setPen(Qt::gray);
        const QuANRect apply_r(x, 0, data->digitrects.last().width(), data->digitrects.last().height(), QuANRect::Apply);
        const QBrush b = p->brush();
        const QFont fo = p->font();
        if(apply_r.contains(data->mouse_pos) && data->data != data->last_applied_data) {
            p->setBrush(QBrush(bgcolor.darker(110)));
            p->drawRect(apply_r);
        }
        data->digitrects.push_back(apply_r);
        // draw an apply tick within the font rect
        QFont f(fo);
        f.setPointSizeF(0.75 * fo.pointSizeF());
        p->setFont(f);
        p->setPen(data->data != data->last_applied_data ? QColor(Qt::green) : QColor(Qt::gray));
        p->drawText(apply_r, CHECK_MARK,  Qt::AlignHCenter|Qt::AlignVCenter);
        p->setBrush(b); // restore brush
        p->setFont(fo); // restore font
    }


    p->setPen(Qt::gray);
    p->drawRoundedRect(rect, 1, 1);

}

void QuNumericEngine::m_drawHighlighted() {

}

// returns the index of the digit: right most digit exp 0
int QuNumericEngine::m_map_to_digit_pos(const QPointF &mousePos) const {
    for(int i = 0; i < data->digitrects.size(); i++) {
        if(data->digitrects[i].contains(mousePos) && data->digitrects[i].pos >= 0) {
            return data->digits - data->digitrects[i].pos - 1;
        }
        else if(data->digitrects[i].contains(mousePos))
            return data->digitrects[i].pos; // "+" (pos is negative)
    }
    return QuANRect::Invalid; // "+": pos is < 0
}

bool QuNumericEngine::m_up(const QPointF &mousePos) const {
    if(data->digitrects.size() > 0)
        return mousePos.y() < data->digitrects[0].height() / 2.0;
    return false;
}
void QuNumericEngine::setFont(const QFont &f) {
    data->font = f;
    data->font.setPointSizeF(f.pointSizeF() * data->font_scale);
    data->arrow_hei = QFontMetrics(data->font).height() / 4.0;
    m_update_maxcharw();
    data->minsiz = QSizeF(); // need update minimumSize
}

QSizeF QuNumericEngine::minimumSize()  {
    if(data->minsiz.isValid())
        return data->minsiz;
    qDebug () << __PRETTY_FUNCTION__ << data->font;
    const QString& sign ="-"; // always reserve space for "-" sign to avoid unwanted resizes
    QFontMetrics fm(data->font);
    int fh = fm.height();
    double minh = fh + 2.1 * data->arrow_hei + data->bm + data->um;
    const int nrects = data->show_apply ? data->digits + 1 : data->digits;
    data->minsiz = QSizeF (nrects * data->maxcharw + fm.horizontalAdvance(sign)
               + (data->decDig > 0 ? fm.horizontalAdvance(".") : 0) + data->lm + data->rm, minh);
    qDebug() << __PRETTY_FUNCTION__ << data->font << "pt. siz" << data->font.pointSizeF() << "ascent + descent" << fh
             << "max w" << data->maxcharw << "size width" << data->minsiz.width() ;
    return  data->minsiz;
}

void QuNumericEngine::setValue(double v)
{
    long long temp = (long long) round(v * pow(10.0, data->decDig));
    if ((temp >= data->min) && (temp <= data->max))
    {
        bool valChanged = data->data != temp;
        data->data = temp;
        if (valChanged) {
            emit valueChanged(value());
        }
    }
    else
        perr("%s: value %f out of range [%f, %f] ( %lld [%lld, %lld])",
             __PRETTY_FUNCTION__, v, data->min_d, data->max_d, temp, data->min, data->max);
}

void QuNumericEngine::setMaximum(double m) {
    if (m >= data->min_d && m != data->max_d) {
        data->max_d = m;
        printf("CuApplyNumericEngine::setMaximum to %f. internally from %lld ", m, data->max);
        data->max = (long long) round(m* (long long)pow(10.0, data->decDig));
        printf(" to %lld (%f)\n", data->max, data->max_d);
        emit maximumChanged(data->max_d);
    }
}

void QuNumericEngine::setMinimum(double m) {
    if (m <= data->max_d && m != data->min_d) {
        data->min_d = m;
        printf("CuApplyNumericEngine::setMinimum to %f. internally from %lld ", m, data->min);
        data->min = (long long) round(m* (long long)pow(10.0, data->decDig));
        printf(" to %lld (%f)\n", data->min, data->min_d);
        emit minimumChanged(data->min_d);
    }
}

void QuNumericEngine::setIntDigits(int i)
{
    if (i > 0 && i != data->intDig) {
        data->intDig = i;
        data->digits = data->intDig + data->decDig;
        m_update_bounds();
        data->minsiz = QSizeF(); // need update minimumSize
        // caller needs update()  (paint)
        emit intDigitsChanged(i);
    }
}

void QuNumericEngine::setDecDigits(int dd)
{
    if(dd >= 0  && dd != data->decDig) {
        printf("setDecDigits: dec digits from %d to %d  %lld (%f) ", data->decDig, dd, data->data, value());
        data->data = (long long) (data->data * pow(10.0, dd - data->decDig));
        data->min = (long long) (data->min * pow(10.0, dd - data->decDig));
        data->max = (long long) (data->max * pow(10.0, dd - data->decDig));
        data->decDig = dd;
        data->digits = data->intDig + data->decDig;
        printf("to \e[1;32m%lld (%f)\e[0m\n", data->data, value());
        m_update_bounds();
        // caller needs update()  (paint)
        data->minsiz = QSizeF(); // need update minimumSize
        emit decDigitsChanged(dd);
    }
}

void QuNumericEngine::click(const QPointF& pos) {
    bool up = m_up(pos);
    click(pos, up);
}

void QuNumericEngine::click(const QPointF &pos, bool up) {
    printf("CuApplyNumericEngine::click \e[1;35m %lld (%f) \e[1;34m ... ", data->data, value());
    if(!pos.isNull()) {
        int dig = m_map_to_digit_pos(pos);
        printf(" Dig is %d ", dig);
        if (dig > -1 && up && (data->data + pow(10.0, dig)) <= data->max) {
            data->data += (long long) pow(10.0, dig);
            emit valueChanged(value());
        }
        else if(dig > -1 && !up && data->data - pow(10.0, dig) >= data->min) {
            data->data -= (long long) pow(10.0, dig);
            emit valueChanged(value());
        }
        else if(m_rect_type(dig) == QuANRect::Apply) {
            data->last_applied_data = data->data;
            emit applyClicked(value());
        }
    }
    printf("\e[1;32m to %lld (%f)\e[0m\n", data->data, value());
}

void QuNumericEngine::setShowApply(bool show) {
    data->show_apply = show;
    m_update_maxcharw();
    data->minsiz = QSizeF(); // need update minimumSize
}

void QuNumericEngine::m_update_bounds()  {
    bool c;
    long long int m = (long long int) -pow(10.0, data->digits) + 1;
    long long int M =  (long long int) pow(10.0, data->digits) - 1;
    printf("\e[0;33mm_update_bounds old m %lld --> %lld \e[0;32m old M %lld --> %lld\e[0m", data->min, m, data->max, M);
    if(m > data->min) {
        data->min = m;
        data->min_d = -pow(10.0, data->intDig) + pow(10.0, -data->decDig);
        emit minimumChanged(data->min_d);
    }
    if(M < data->max) {
        data->max = M;
        data->max_d = pow(10.0, data->intDig) - pow(10.0, -data->decDig);
        emit maximumChanged(data->max_d);
    }
    printf("\e[0;36m %lld < %lld < %lld \e[1;35m %f < %f < %f\e[0m\n", data->min, data->data, data->max, data->min_d, value() , data->max_d);
    c = data->data < data->min || data->data > data->max;
    if(data->data < data->min)
        data->data = data->min;
    if(data->data > data->max)
        data->data = data->max;
    if(c) {
        emit valueChanged(value());
        printf("\e[1;31mdata changed to %lld ( %f ) due to bounds changed \e[0m\n", data->data, value());
    }
    printf("with %d int digs and %d dec digs: \e[1;35mm_update_bounds %.3f - %.3f\e[0m\n", data->intDig, data->decDig, data->min_d, data->max_d);
}

double QuNumericEngine::m_to_double(long long int lli) const {
    return lli * pow(10.0, -data->decDig);
}

void QuNumericEngine::m_update_maxcharw() {
    QFontMetrics fm(data->font);
    data->maxcharw = 0;
    double w;
    for(int i = 0; i < 10; i++) {
        w = fm.horizontalAdvance(QString("%1").arg(i));
        if(data->maxcharw < w)
            data->maxcharw = w;
    }
    if(data->show_apply && data->maxcharw < fm.horizontalAdvance(CHECK_MARK))
        data->maxcharw = fm.horizontalAdvance(CHECK_MARK);
}

QuANRect::Type QuNumericEngine::m_rect_type(int idx) const {
    if(idx >= 0)
        return QuANRect::Digit;
    return static_cast<QuANRect::Type> (idx);
}

