#ifndef DRAWABLE_I_H
#define DRAWABLE_I_H

#include <QPointF>
#include <QSizeF>

class QRectF;
class QPainter;

/*!
 * \brief The Drawable_I is an interface for any object that shall be drawn into a given rectangle
 *
 * The Drawable_I does not define *where* the *drawable* is drawn (QWidget, QGraphicsItem, ...)
 */
class Drawable_I {
public:

    /*!
     * \brief returns true if the drawable scales with zooming
     *
     * \return true the drawable scales according to the zoom. It will be painted with the painter *transform* applied
     *         false the drawable does not scale with the zoom. It will be painted after the painter is *restored* or
     *         before the transformation is applied, according to the z value
     * \see z
     */
    virtual bool scales() = 0;

    /*!
     * \brief z defines the z order of the element that is drawn
     * \return the z order
     *
     * Drawable elements with a lower z value are drawn before elements with a higher z.
     */
    virtual int z() const = 0;

    /*!
     * \brief subclasses will draw themselves using the provided painter within
     *        the area in the given rect
     * \param p pointer to the painter
     * \param r the available area
     *
     * If relativeRect returns a valid rect, the *drawable* shall draw itself
     * within a sub rectangle of r such as
     *
     * \code
      const QRectF& rel = relativeRect();
      QRectF drawR = QRectF(r.width() * rel.x(),
           r.height() * rel.y(),
           r.width() * rel.width(),
           r.height() * rel.height());
      // draw inside drawR
     * \endcode
     */
    virtual void draw(QPainter *p, const QRectF &r) = 0;
};

#endif // DRAWABLE_I_H
